@@ -281,28 +281,44 @@ static PyObject *
 Reader_write(Reader *self, PyObject *args, PyObject *kwds)
 {
     char* epc_data;
-    char* epc_target;
+    char* epc_target = NULL;
     TMR_Status ret;
     TMR_TagData data;
     TMR_TagData target;
-    TMR_TagFilter filter;
+    TMR_TagFilter tag_filter, *filter;
+
     // Read call arguments.
-    static char *kwlist[] = {"epc_target", "epc_code", NULL};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "ss", kwlist, &epc_target, &epc_data))
+    static char *kwlist[] = {"epc_code", "epc_target", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|s", kwlist, &epc_data, &epc_target))
         return NULL;
-    /* build target tag to search */
-    target.epcByteCount = strlen(epc_target) * sizeof(char) / 2;
-    TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
+
     /* build data tag to be writen */
     data.epcByteCount = strlen(epc_data) * sizeof(char) / 2;
     TMR_hexToBytes(epc_data, data.epc, data.epcByteCount, NULL);
-    // Build filter target tag to be replaced.
-    TMR_TF_init_tag(&filter, &target);
+
+    /* build target tag to search */
+    if(epc_target != NULL)
+    {
+        target.epcByteCount = strlen(epc_target) * sizeof(char) / 2;
+        TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
+
+        filter = &tag_filter;
+        TMR_TF_init_tag(filter, &target);
+    }
+    else
+        filter = NULL;
+
     // Write data tag on target tag.
-    ret = TMR_writeTag(&self->reader, &filter, &data);
+    ret = TMR_writeTag(&self->reader, filter, &data);
     // In case of not target tag found.
     if (ret == TMR_ERROR_NO_TAGS_FOUND)
         Py_RETURN_FALSE;
+    else if(ret != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
     Py_RETURN_TRUE;
 }
 
@@ -487,6 +503,91 @@ Reader_stop_reading(Reader* self)
     Py_RETURN_NONE;
 }
 
+static PyObject *
+Reader_read_tag_mem(Reader *self, PyObject *args, PyObject *kwds)
+{
+    TMR_Status ret;
+    char* epc_target = NULL;
+    uint32_t bank, address, count;
+
+    TMR_TagData target;
+    TMR_TagFilter tag_filter, *filter;
+    uint8_t *buf;
+    PyObject *result;
+
+    static char *kwlist[] = {"bank", "address", "count", "epc_target", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "III|s", kwlist, &bank, &address, &count, &epc_target))
+        return NULL;
+
+    if(epc_target != NULL)
+    {
+        /* build target tag to search */
+        target.epcByteCount = strlen(epc_target) / 2;
+        TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
+
+        filter = &tag_filter;
+        TMR_TF_init_tag(filter, &target);
+    }
+    else
+        filter = NULL;
+
+    buf = malloc(count);
+    ret = TMR_readTagMemBytes(&self->reader, filter, bank, address, (uint16_t)count, buf);
+    if (ret == TMR_ERROR_NO_TAGS_FOUND)
+        Py_RETURN_NONE;
+    else if (ret != TMR_SUCCESS)
+    {
+        free(buf);
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    result = PyByteArray_FromStringAndSize((const char *)buf, count);
+    free(buf);
+
+    return result;
+}
+
+static PyObject *
+Reader_write_tag_mem(Reader *self, PyObject *args, PyObject *kwds)
+{
+    TMR_Status ret;
+    char* epc_target = NULL;
+    uint32_t bank, address;
+    PyObject *data;
+
+    TMR_TagData target;
+    TMR_TagFilter tag_filter, *filter;
+
+    static char *kwlist[] = {"bank", "address", "data", "epc_target", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "IIO!|s", kwlist, &bank, &address, &PyByteArray_Type, &data, &epc_target))
+        return NULL;
+
+    if(epc_target != NULL)
+    {
+        /* build target tag to search */
+        target.epcByteCount = strlen(epc_target) / 2;
+        TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
+
+        filter = &tag_filter;
+        TMR_TF_init_tag(filter, &target);
+    }
+    else
+        filter = NULL;
+
+    ret = TMR_writeTagMemBytes(&self->reader, filter, bank, address, PyByteArray_Size(data), (uint8_t *)PyByteArray_AsString(data));
+    free(filter);
+    if (ret == TMR_ERROR_NO_TAGS_FOUND)
+        Py_RETURN_FALSE;
+    else if (ret != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+    else
+        Py_RETURN_TRUE;
+}
+
 static PyObject *
 Reader_gpi_get(Reader *self, PyObject *args)
 {
@@ -502,7 +603,7 @@ Reader_gpi_get(Reader *self, PyObject *args)
 
     if ((ret = TMR_gpiGet(&self->reader, &count, gpio)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         free(gpio);
         return NULL;
     }
@@ -525,20 +626,29 @@ static PyObject *
 Reader_gpo_set(Reader *self, PyObject *args)
 {
     uint8_t pin;
-    bool value;
+    PyObject *value;
+    int intval;
     TMR_GpioPin gpio;
     TMR_Status ret;
 
-    if (!PyArg_ParseTuple(args, "Bp", &pin, &value))
+    /* to preserve Python 2.x compatibility we cannot use "p"
+       also, using "i" is foolish as per https://bugs.python.org/issue14705 */
+    if (!PyArg_ParseTuple(args, "BO", &pin, &value))
         return NULL;
 
+    if ((intval = PyObject_IsTrue(value)) < 0)
+    {
+        PyErr_SetString(PyExc_TypeError, "Boolean value expected");
+        return NULL;
+    }
+
     gpio.id = pin;
-    gpio.high = value;
+    gpio.high = !!intval; /* convert to bool */
     gpio.output = true;
 
     if ((ret = TMR_gpoSet(&self->reader, 1, &gpio)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -559,7 +669,7 @@ Reader_get_model(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_VERSION_MODEL, &model)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -633,7 +743,7 @@ Reader_set_region(Reader *self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_REGION_ID, &region)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -656,7 +766,7 @@ Reader_get_supported_regions(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_REGION_SUPPORTEDREGIONS, &regions)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -687,7 +797,7 @@ get_uint32(TMR_Reader *reader, int param)
 
     if ((ret = TMR_paramGet(reader, param, &value)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         return NULL;
     }
 
@@ -705,7 +815,7 @@ set_uint32(TMR_Reader *reader, PyObject *args, int param)
 
     if ((ret = TMR_paramSet(reader, param, &value)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         return NULL;
     }
 
@@ -726,7 +836,7 @@ get_uint8List(TMR_Reader *reader, int param, size_t max)
 
     if ((ret = TMR_paramGet(reader, param, &values)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         free(values.list);
         return NULL;
     }
@@ -756,7 +866,7 @@ get_uint32List(TMR_Reader *reader, int param, size_t max)
 
     if ((ret = TMR_paramGet(reader, param, &values)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         free(values.list);
         return NULL;
     }
@@ -797,7 +907,7 @@ set_uint8List(TMR_Reader *reader, PyObject *args, int param)
 
     if ((ret = TMR_paramSet(reader, param, &values)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         free(values.list);
         return NULL;
     }
@@ -831,7 +941,7 @@ set_uint32List(TMR_Reader *reader, PyObject *args, int param)
 
     if ((ret = TMR_paramSet(reader, param, &values)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         free(values.list);
         return NULL;
     }
@@ -898,14 +1008,14 @@ Reader_get_power_range(Reader *self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_RADIO_POWERMIN, &lim_power)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
     PyTuple_SetItem(powers, 0, PyLong_FromLong(lim_power));
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_RADIO_POWERMAX, &lim_power)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
     PyTuple_SetItem(powers, 1, PyLong_FromLong(lim_power));
@@ -928,7 +1038,7 @@ get_PortValueList(TMR_Reader *reader, int param)
 
     if ((ret = TMR_paramGet(reader, param, &value_list)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         return NULL;
     }
 
@@ -1001,7 +1111,7 @@ set_PortValueList(TMR_Reader *reader, PyObject *args, int param)
     Py_DECREF(input);
     return get_PortValueList(reader, param);
 fail:
-    PyErr_SetString(PyExc_TypeError, TMR_strerr(reader, ret));
+    PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
     return NULL;
 }
 
@@ -1062,7 +1172,7 @@ Reader_get_gen2_q(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_Q, &model)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1085,7 +1195,7 @@ Reader_set_gen2_q(Reader* self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_Q, &model)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1104,7 +1214,7 @@ Reader_get_gen2_tagencoding(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_TAGENCODING, &tagencoding_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1122,7 +1232,7 @@ Reader_set_gen2_tagencoding(Reader* self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_TAGENCODING, &tagencoding_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1137,7 +1247,7 @@ Reader_get_gen2_session(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_SESSION, &session_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1155,7 +1265,7 @@ Reader_set_gen2_session(Reader* self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_SESSION, &session_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1170,7 +1280,7 @@ Reader_get_gen2_target(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_TARGET, &target_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1188,7 +1298,7 @@ Reader_set_gen2_target(Reader* self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_TARGET, &target_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1203,7 +1313,7 @@ Reader_get_gen2_blf(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_BLF, &blf_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1221,7 +1331,7 @@ Reader_set_gen2_blf(Reader* self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_BLF, &blf_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1236,7 +1346,7 @@ Reader_get_gen2_tari(Reader* self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_TARI, &tari_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1254,7 +1364,7 @@ Reader_set_gen2_tari(Reader* self, PyObject *args)
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_TARI, &tari_val)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1269,7 +1379,7 @@ Reader_get_temperature(Reader *self)
 
     if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_RADIO_TEMPERATURE, &temp)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_TypeError, "Error getting temperature");
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
         return NULL;
     }
 
@@ -1292,14 +1402,18 @@ static PyMethodDef Reader_methods[] = {
     {"stop_reading", (PyCFunction)Reader_stop_reading, METH_NOARGS,
      "Stop asynchronous reading"
     },
+    {"read_tag_mem", (PyCFunction)Reader_read_tag_mem, METH_VARARGS | METH_KEYWORDS,
+     "Read bytes from the memory bank of a tag"
+    },
+    {"write_tag_mem", (PyCFunction)Reader_write_tag_mem, METH_VARARGS | METH_KEYWORDS,
+     "Write bytes to the memory bank of a tag"
+    },
     {"gpi_get", (PyCFunction)Reader_gpi_get, METH_VARARGS,
      "Gets GPIO pin value"
     },
-#if PY_MAJOR_VERSION >= 3
     {"gpo_set", (PyCFunction)Reader_gpo_set, METH_VARARGS,
      "Sets GPIO pin value"
     },
-#endif
     /* Reader parameters */
     {"get_model", (PyCFunction)Reader_get_model, METH_NOARGS,
      "Returns the model name"
